'''
File:    csv_animate.py
By:      Jerry Farm
Created: 2005.11.06
Revised: 2006.01.05

This Python script will import a .CSV file (comma separated values, e.g.
as saved from Microsoft Excel) and save the data to Blender Ipo objects.
In this way, time history data generated elsewhere can be brought into
Blender and used to define animation motions.

This file is intended to be generic and actually contains several functions
that are to be called from another script created by the user.  That script
manages the details specific to their application, namely, which columns
get mapped to which curves of which objects.

YOU MUST DOWNLOAD AND INSTALL PYTHON FOR THIS SCRIPT TO WORK.
http://python.org
This script will not work with just the basic Blender install.
Python 2.3 or later is required for its csv file reader.


Functions - these are the functions the user's script will call:

csv_animate.read_file(csv_file)
  Reads the specified csv file into memory.  Since I long-ago tired of
  counting column numbers, the csv file must have a header, i.e., the first
  line of the file must contain comma-separated column names, with no quotes
  or spaces.  This function stores each column of the the csv file as a list
  object in the global name space using the corresponding name from the header.

  Parameters:
    csv_file - a string with the csv file name (and path, if needed)

csv_animate.set_frames ()
  Sets Blender's animation start frame to 1 and end frame to the number of 
  frames read in by read_file().
  
csv_animate.set_timing ()
  Sets Blender's animation frames/second (fps) to 1/dt, where dt is the
  time interval of the data read in by read_file().  This assumes that the
  csv file contains a column labeled "t" (without the quotes), and that the
  time interval is the same throughout the file.  The time interval is 
  computed as dt = t[1] - t[0]
  Do not call this function if the csv file does not have a time vector.
  In this case the user must set fps either with their script or 
  manually through the Blender GUI,

csv_animate.set_extension (name_extension)
  Sets an extension string that csv_animate.animate() concatenates to the 
  object_name and ipo_name.  Useful inside a loop when animating multiple
  objects which have the same name but different numbers,
  e.g. "Cube.001", "Cube.002".
  Initially set to an empty string.
  
  Parameters:
    name_extension - an integer or string defining the name extension.
      If a string is used it must include the leading dot, e.g. ".001"

csv_animate.animate (object_name, ipo_name, curves)
  Saves data as keyframes to a Blender Ipo object to animate a Blender Object.
  The extension set by csv_animate.set_extension() will be appended
  internally to object_name and ipo_name before they are used to call Blender.

  Parameters:

    object_name - a string with the name of the Blender Object to be animated.
      Only one call to animate() can be made per Object--subsequent calls will
      discard data from previous calls.

    ipo_name - a string with the name of the Ipo to create for the Object

    curves - a list specifying all the data to use for animating the Object.
      Each entry in the list is a tuple defining one aspect of the animation
      for the Object.  Each tuple must be in one of the following forms:
      Form 1:   (curvename, data)
      Form 2:   ("Axis", data, rotation_axis)

      Form 1:  (curvename, data)
      Defines data for one Ipo curve
          curvename - a string containing one of the valid Ipo curve names,
                      e.g. "LocX", "RotY", "SizeZ"
          data - can be either a list of data values, or a string
                 list - contains the values to be stored into the Ipo curve.
                   Does not have to be the same length as the number of frames.
                 string - contains a valid Python expression to be used
                   to compute the values for the curve.  The expression is
                   evaluated within the scope of csv_animate, and will be
                   evaluated once for each row of the csv file (once for each
                   frame of the animation).  Normally you access the variables
                   created by read_file() from the columns of csv file.
                   The variable 'i' is available for use in the expression and
                   holds the index into the csv data for that evaluation
                   (starting at 0 since it indexes lists).  This provides 
                   an easy, shorthand method for doing simple calculations on
                   the data without needing to run a separate loop beforehand.

      Form 2:  ("Axis", data, rotation_axis)
      Values for RotX, RotY, RotZ will be computed to give a right-hand
      rotation of (data) degrees about the axis (rotation_axis)
          "Axis" - case sensitive, differentiates this tuple from Form 1.
          data   - list or string defining rotation angles (see above) 
          rotation_axis - a 3-tuple with a vector defining the axis of rotation.
              The axis is time-invariant.  The vector need not be normalized.

'''
###############################################################################

import Blender
from Blender import Object, Ipo, Camera, Scene
from Blender.Mathutils import *
from Blender.Scene import Render

import csv

# Import this in case used by input file
import math
from math import *


# Public variables

nframes = 0
t   = []  # Time vector from csv file (empty list if none)
t0  = 0   # Start time (first entry in t, 0 if no t)
dt  = 1   # Time interval (t[1] - t[0], 1 if no t)
fps = 1   # Frames per second (1/dt, t if no t)

name_extension_str = ""   # Object name extension


#------------------------------------------------------------------------------
# Internal Functions
#------------------------------------------------------------------------------

# Add to list subscript (starts at 0) to convert to keyframe index (starts at 1)
ioffset = 1

#------------------------------------------------------------------------------

def new_object_ipo (object_name, ipo_name):

	try:
		obj = Blender.Object.Get(object_name)
	except:
		print
		print "* * * No object named", object_name
		print
		raise

	ipo = obj.getIpo()
	if ipo:
		# Already have an Ipo--clear its curves and rename it
		print ipo
		curves = ipo.getCurves()
		print curves
		for curve in curves:
			print "Curve", curve
			curve_name = curve.getName()
			print "Removing", curve_name
			ipo.delCurve(curve_name)

		ipo.setName(ipo_name)

		'''
		# (In Blender 2.37 delCurve has a bug that can cause a crash.
		# use this section instead the above for versions before 2.40)
		# Already have an Ipo--trash it and make a new one
		print "Clearing existing Ipo",ipo.getName(), "from", object_name
		obj.clearIpo()   # unlinks Ipo from object
		# If this script is re-run on the same model without restarting
		# Blender, it seems to slowly leak memory, presumably due
		# to the curves not being deleted (see above)--or else I haven't
		# done it enough for garbage collection to happen.
		# Since this script normally isn't run that often,
		# I'm not going to worry about it for now.
		ipo = Ipo.New('Object',ipo_name)
		obj.setIpo(ipo)
		'''
	else:
		ipo = Ipo.New('Object',ipo_name)
		obj.setIpo(ipo)

	print object_name, " now has Ipo ", ipo.getName()

	return (ipo)

#------------------------------------------------------------------------------

def add_curve (ipo, curve_name, curve_data):

	c = ipo.addCurve(curve_name)
	c.setInterpolation('Linear')
	c.setExtrapolation('Constant')
	print '  ', curve_name, ' created for ', ipo.getName()

	if curve_name.find("Rot") > -1:
		# This is a rotation (e.g. RotX or dRotX)
		# Convert to Ipo units (1 unit = 10 deg)
		scale = 0.1
	else:
		scale = 1.0

	if isinstance (curve_data, str):
		# curve_data is a string, execute it
		for i in range(nframes):
			a = eval(curve_data) * scale
			c.addBezier( (i+ioffset, a) )
	else:
		# curve_data is a vector, index it
		for i in range(len(curve_data)):
			a = curve_data[i] * scale
			c.addBezier( (i+ioffset, a) )

	# Update Bezier parameters
	# This seems to mess up if this script is run twice on the same model.
	# Since we are defining every frame as a key frame and won't be
	# interpolating between frames, we don't seem to need this anyway, 
	# so leave it out.
	# (Potential for problems on interlaced or motion blur if they
	# render "between frames"?  Or, maybe by using Linear Interpolation,
	# recalc() is not needed and there is no problem there either.)
	#c.recalc()

#------------------------------------------------------------------------------

def add_axis (ipo, curve_data, axis):

	# Prepare the rotation angles
	rots = []
	if isinstance (curve_data, str):
		# curve_data is a string, execute it
		for i in range(nframes):
			rots.append( eval(curve_data) )
	else:
		# curve_data is a vector, index it
		for i in range(len(curve_data)):
			rots.append( curve_data[i] )

	# Compute the Euler angles for all frames
	angles = []
	for i in range(len(rots)):
		v = Vector(axis)
		rmat = RotationMatrix(rots[i], 3, 'r', v)
		euler_angles  = rmat.toEuler()
		angles.append([euler_angles.x, euler_angles.y, euler_angles.z])

	# Copy Euler angles to Ipo rotation curves
	# For speed, we process all frames on a curve at once
	# (rather than process all curves in a frame)
	rot_names = ["RotX","RotY","RotZ"]
	for icurve in range(len(rot_names)):
		rot_name = rot_names[icurve]
		c = ipo.addCurve(rot_name)
		c.setInterpolation('Linear')
		c.setExtrapolation('Constant')
		print '  ', rot_name, ' created for ', ipo.getName()
		for i in range(len(rots)):
			# note angles divided by 10 for Ipo rotation
			c.addBezier( (i+ioffset,angles[i][icurve]/10.0) )
		#c.recalc()

#------------------------------------------------------------------------------
# External Functions
#------------------------------------------------------------------------------

def read_file (csv_file):

	f = open(csv_file,'rb')
	reader = csv.reader(f)

	ncolumns = 0
	columns  = []
	column_names = []
	
	global nframes
	nframes = 0

	# Read the csv file, splitting the data out by columns
	# The file must start with a row of column names
	# The names must be valid variable names

	for row in reader:

		if not column_names:

			# Don't have column names yet, 
			# so save this first row as the column names
			column_names = row
			ncolumns = len(column_names)
			print "Column names are"
			print column_names
			print "Importing data..."

			# Create a list of empty columns to collect data
			for i in range(ncolumns):
				columns.append([])

		else:

			# Tack new data to end of each column
			for i in range(ncolumns):
				columns[i].append(eval(row[i]))

			nframes = nframes + 1
			if nframes % 100 == 0:
				print "Frame ", nframes

	f.close()  # Be kind, rewind


	print
	print nframes, "frames imported"
	print "Assigning data columns to separate variables..."

	# Break out the columns into named variables
	for i in range(ncolumns):
		estr = column_names[i] + " = columns["+str(i)+"]"
		print estr  # display the assignment
		# tack on global so variable is visible outside of csv_animate
		estr = "global " + column_names[i] +"\n" + estr
		exec estr

#------------------------------------------------------------------------------

def set_extension (name_extension):

	global name_extension_str
	if isinstance (name_extension, str):
		name_extension_str = name_extension
	else:
		name_extension_str = ".%03d" % name_extension

#------------------------------------------------------------------------------

def animate (object_name, ipo_name, curves):

	ipo = new_object_ipo (object_name+name_extension_str, ipo_name+name_extension_str)
	for curve in curves:
		if curve[0] == "Axis":
			add_axis  (ipo, curve[1], curve[2])
		else:
			add_curve (ipo, curve[0], curve[1])

#------------------------------------------------------------------------------

def set_frames ():

	scn = Scene.GetCurrent()
	context = scn.getRenderingContext()

	context.startFrame(1)
	context.endFrame(nframes)

#------------------------------------------------------------------------------

class NoTimeVector(Exception):
	def __str__(self):
		return repr(self.value)

#------------------------------------------------------------------------------

def set_timing ():

	if t:
		global  dt, t0, fps
		t0  = t[0]
		dt  = t[1] - t0
		fps = 1/dt
		scn = Scene.GetCurrent()
		context = scn.getRenderingContext()
		context.framesPerSec(int(round(fps)))
	else:
		print
		print "ERROR:  Cant't set animation timing."
		print "No time vector was loaded."
		raise NoTimeVector

