'''
csv_demo.py
Demonstration of csv_animate

YOU MUST DOWNLOAD AND INSTALL PYTHON FOR THIS SCRIPT TO WORK.
http://python.org
This script will not work with just the basic Blender install.

To use:
1) Start Blender and open csv_demo.blend (or double-click csv_demo.blend)
2) Using the Screen Selector, switch to the scripting screen (SCR:5 - Scripting)
3) In the Text Toolbar (at bottom of Text Window), do File > Open (or ALT-O)
   to open the file csv_demo.py
4) Execute the script (ALT-P)
5) Switch to the Animation screen (SCR:1 - Animation)
6) Play the timeline (View > Play Back Animation, or ALT-A). ESC to stop playback.
   IMPORTANT: It seems you must play the time line or otherwise cause the
   frame number to change before rendering the animation, otherwise the first
   frame of the animation will be rendered without the correct animation data.
   (You'll probably want to hit play anyway to check the animation.)
7) If desired, click ANIM to render the animation to an AVI file.

'''

import csv_animate
from csv_animate import *


# Read in a csv file
csv_file = './csv_demo.csv'
csv_animate.read_file(csv_file)

# read_file() copies the columns of the csv_file to separate variables
# names according to the column headers in the file, thus, we now have:
#   csv_animate.t[]
#   csv_animate.sin[]
#   csv_animate.cos[]
#   csv_animate.quadratic[]

# Set the number of frames in the animation based on the csv data
csv_animate.set_frames ()

# Set the frames/second based on the time vector of the csv data
csv_animate.set_timing ()

# Make a handy local copy of the number of frames
nframes = csv_animate.nframes



# Example:  Setting a constant Ipo curve value
# "Base" is a 2x2x2 cube, centered at origin.
# Move it up 1 unit so it is sitting on the Plane.
# Also move it -1 unit on X.
# Since the Base is already at Y=0 where we want it, don't need to set LocY
# (though we could)
# Note that the Base was made the parent of the rocker, so the rocker will 
# move up with the Base.
# Setting LocX,LocZ will override the original x & z position of the Base.
# When you first start or reset the animation, you'll see the Base jump up
# to this position.

animate ("Base", "BaseIpo",
	[
	("LocX", [-1.0] ),  # the constant must be wrapped in a list
	("LocZ", [1.0] )
	]
)


# Example:  Setting a rotation Ipo curve value
# Use the "sin" column of the csv file to rotate the rocker back and forth.
# The sin column only goes to magnitude +/-1, so scale to +/-45 degrees.
# sin[i] is evaluated, i.e, eval(), inside the animate() function.
# The sin[] variable was created in read_file(), 
# and i is an internal indexing variable inside the animate() function.

animate ("rocker", "rockerIpo",
	[
	("RotX", "sin[i]*45.0" )  # string argument is evaluated inside animate()
	]
)


# Example:  Setting a rotation around and arbitrary axis
# Rotate the spin_cube around a diagonal axis, i.e. an axis through its
# center and through the 1,1,1 corner.  This axis is represented by the
# vector (1,1,1) so pass that in as the rotation axis.  
# Note that the rotation axis need not be normalized.
# To make one rotation in 3 seconds, use an angle of t*360(deg)/3(sec)

animate ("spin_cube", "spin_cubeIpo",
	[
	("Axis", "t[i]*360.0/3.0", (1.0, 1.0, 1.0))
	]
)


# Example: Move, rotate, and size an object.
# Make the red sphere fall off of the Base and roll on the ground.
#
# Start on top of "south east" corner of Base (x=0, y=-1, z=+2)
# Move east along y=-1, so set y=-1
# Also, moving east at 1 unit per second, starting at x=0,
# means we set x = t.
#
# The fall happens on the z axis, so use the "quadratic" column of the csv
# file to animate z.  The quadratic column goes from 3.0 down to zero.
# But we want the ball to fall from 2.5 (Base size of 2 + 0.5 ball radius) 
# down to 0.5 (ball radius), so scale quadratic by 2/3 and add 0.5.
# Note the float numbers (2.0/3.0) are used to avoid using integer arithmetic
# which will cause discrete jumps in the data.
#
# A ball of radius 0.5 moving at 1 unit/sec rolls at 360*1/(2*0.5pi) deg/sec
# so apply a rotation on Y of t*360/pi
#
# Finally, the sphere has radius 1, but we want 0.5, so apply 0.5 Size to it.
#
# When you first start or reset the animation, you'll see the sphere shrink
# and jump up to the top of the Base.

animate ("Sphere", "sphereIpo",
	[
	("LocX", csv_animate.t ),  # csv_animate.t is equivalent to "t[i]"
	("LocY", "-1.0" ),
	("LocZ", "quadratic[i]*2.0/3.0 + 0.5" ),
	("RotY", "t[i]*360.0/3.14159" ),
	("SizeX", [0.5] ),
	("SizeY", [0.5] ),
	("SizeZ", [0.5] )
	]
)


# Example: Building your own animation data
# If you only pass list data into animate() you don't need to read a csv file.
#
# Make the blue sphere fall off of the Base and bounce, with squish.
# The Bouncer is already radius 0.5, so the only sizing will be for the squish.
#
# Start on top of "south west" corner of Base (x=-2, y=-1, z=+2)
# Move south along x=-2
z = 2.5 # initialize z location
vz = 0  # initialize velocity along z
az = -1 # set acceleration along z
LocZ_list  = []
SizeX_list = []
SizeZ_list = []
for i in range(nframes):
	LocZ_list.append(z)
	# check height against the sphere radius for squish
	if z >= 0.5:
		sz = 1.0
		sx = 1.0
	else:
		sz = 2.0*z			# shrink z
		sx = 1.0 + (0.5 - z)/0.25	# fatten x,y
	SizeX_list.append(sx)
	SizeZ_list.append(sz)
	#print i, z, sx, sz	# echo data
	vz = vz + az * csv_animate.dt
	z  = z  + vz * csv_animate.dt
	# do bounce
	# check height against half the sphere radius to allow squish
	if z < 0.25:
		z  = 0.5
		vz = -0.5*vz
# end loop, now have LocZ_list, SizeX_list, SizeZ_list constructed

# Because the directions for sizing change with the rotation of the ball,
# apply the sizing to a parent lattice which does not rotate.
# Resizing the lattice will in turn deform the ball.
# Since the lattice is the parent, the Location is applied to it as well
animate ("Bouncer_Lattice", "bouncerLIpo",
	[
	("LocX",  [-2.0] ),
	("LocY",  "-1.0-t[i]" ),
	("LocZ",  LocZ_list ),
	("SizeX", SizeX_list ),
	("SizeY", SizeX_list ),  # reuse X data for Y
	("SizeZ", SizeZ_list )
	]
)

# That just leaves the rotation to apply to the ball
animate ("Bouncer", "bouncerIpo",
	[
	("RotX", "t[i]*360.0/3.14159" ),
	]
)



# Example:  Handling duplicate objects
# Duplicates of objects typically have a number like .001 appended to the end.
# These can be handled in a convenient loop using set_extension()

# Alternate string-based approach
#extensions = [".001", ".002"]
#for i,ext in enumerate(extensions):
#	csv_animate.set_extension (ext)

istart = 1
iend   = 2
for i in range(istart, iend+1):

	# Set the name extension to be used by subsequent calls to animate()
	csv_animate.set_extension (i)

	# Position the base
	animate ("Base", "BaseIpo",
		[
		("LocX", [-4.0+i*2] ),  # use loop index to compute position
		("LocY", [4.0] ),
		("LocZ", [1.0] )
		]
	)

	# And set the rocker a'rockin'
	# Note: the i in sin[i] is not the same i as above,
	# because sin[i] is evaluated inside animate()
	# Expressions can be elaborate, using any number of the columns
	# read in by read_file().  Here we use sin[] twice so these rockers
	# only go to one side, but at twice the frequency of the original.
	animate ("rocker", "rockerIpo",
		[
		("RotX", "sin[i]*sin[i]*45.0" )
		]
	)


# Clear extension so it isn't laying around to trip on later.
csv_animate.set_extension ("")


print
print "***** DEMO COMPLETE *****"
print nframes, "frames imported"

